/*
thot package for statistical machine translation
Copyright (C) 2013 Daniel Ortiz-Mart\'inez

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

//--------------- uiPairHashF class: Hash function for pair of unsigned
//                                   int

#include <limits.h>
#include <utility>

class uiPairHashF
{
public:
  enum
  {
    bucket_size = 1
  };

  std::size_t operator()(const std::pair<unsigned int, unsigned int>& pui) const
  {
    return (std::size_t)pui.first * pui.second;
  }

  bool operator()(const std::pair<unsigned int, unsigned int>& left,
                  const std::pair<unsigned int, unsigned int>& right) const
  {
    return left < right;
  }
};
